@@ -99,11 +99,6 @@ static char cPickleCache_doc_string[] =
 #include <stddef.h>
 #undef Py_FindMethod
 
-/* Python 2.4 backward compat */
-#if PY_MAJOR_VERSION <= 2 && PY_MINOR_VERSION < 5
-#define Py_ssize_t int
-typedef Py_ssize_t (*lenfunc)(PyObject *);
-#endif
 
 /* Python string objects to speed lookups; set by module init. */
 static PyObject *py__p_changed;
@@ -111,7 +106,7 @@ static PyObject *py__p_deactivate;
 static PyObject *py__p_jar;
 static PyObject *py__p_oid;
 
-static cPersistenceCAPIstruct *capi;
+static cPersistenceCAPIstruct *cPersistenceCAPI;
 
 /* This object is the pickle cache.  The CACHE_HEAD macro guarantees
    that layout of this struct is the same as the start of
@@ -198,15 +193,29 @@ scan_gc_items(ccobject *self, int target, Py_ssize_t target_bytes)
     */
     insert_after(&before_original_home, self->ring_home.r_prev);
     here = self->ring_home.r_next;   /* least recently used object */
+    /* All objects should be deactivated when the objects count parameter
+     * (target) is zero and the size limit parameter in bytes(target_bytes)
+     * is also zero.
+     *
+     * Otherwise the objects should be collect while one of the following
+     * conditions are True:
+     *  - the ghost count is bigger than the number of objects limit(target).
+     *  - the estimated size in bytes is bigger than the size limit in
+     *    bytes(target_bytes).
+     */
     while (here != &before_original_home &&
-            (self->non_ghost_count > target
-            || (target_bytes && self->total_estimated_size > target_bytes)
+            (
+             (!target && !target_bytes) ||
+             (
+              (target && self->non_ghost_count > target) ||
+              (target_bytes && self->total_estimated_size > target_bytes)
+             )
             )
           )
     {
         assert(self->ring_lock);
         assert(here != &self->ring_home);
-        
+
         /* At this point we know that the ring only contains nodes
             from persistent objects, plus our own home node.  We know
             this because the ring lock is held.  We can safely assume
@@ -214,14 +223,14 @@ scan_gc_items(ccobject *self, int target, Py_ssize_t target_bytes)
             is not the home */
         object = OBJECT_FROM_RING(self, here);
 
-        if (object->state == cPersistent_UPTODATE_STATE) 
+        if (object->state == cPersistent_UPTODATE_STATE)
         {
             CPersistentRing placeholder;
             PyObject *method;
             PyObject *temp;
             int error_occurred = 0;
             /* deactivate it. This is the main memory saver. */
-            
+
             /* Add a placeholder, a dummy node in the ring.  We need
                 to do this to mark our position in the ring.  It is
                 possible that the PyObject_GetAttr() call below will
@@ -237,7 +246,7 @@ scan_gc_items(ccobject *self, int target, Py_ssize_t target_bytes)
             method = PyObject_GetAttr((PyObject *)object, py__p_deactivate);
             if (method == NULL)
                 error_occurred = 1;
-            else 
+            else
             {
                 temp = PyObject_CallObject(method, NULL);
                 Py_DECREF(method);
@@ -523,7 +532,7 @@ cc_debug_info(ccobject *self)
             v = Py_BuildValue("Oi", k, v->ob_refcnt);
 
         else if (! PyType_Check(v) &&
-                (v->ob_type->tp_basicsize >= sizeof(cPersistentObject))
+                PER_TypeCheck(v)
                 )
             v = Py_BuildValue("Oisi",
                             k, v->ob_refcnt, v->ob_type->tp_name,
@@ -605,54 +614,60 @@ cc_oid_unreferenced(ccobject *self, PyObject *oid)
         not release the global interpreter lock until this is
         complete. */
 
-    PyObject *v;
+    cPersistentObject *dead_pers_obj;
+    ccobject* dead_pers_obj_ref_to_self;
 
     /* If the cache has been cleared by GC, data will be NULL. */
     if (!self->data)
         return;
 
-    v = PyDict_GetItem(self->data, oid);
-    assert(v);
-    assert(v->ob_refcnt == 0);
+    dead_pers_obj = (cPersistentObject*)PyDict_GetItem(self->data, oid);
+    assert(dead_pers_obj);
+    assert(dead_pers_obj->ob_refcnt == 0);
+
+    dead_pers_obj_ref_to_self = (ccobject*)dead_pers_obj->cache;
+    assert(dead_pers_obj_ref_to_self == self);
+
     /* Need to be very hairy here because a dictionary is about
         to decref an already deleted object.
     */
 
-#ifdef Py_TRACE_REFS
-    /* This is called from the deallocation function after the
-        interpreter has untracked the reference.  Track it again.
-    */
-    _Py_NewReference(v);
-    /* Don't increment total refcount as a result of the
-        shenanigans played in this function.  The _Py_NewReference()
-        call above creates artificial references to v.
-    */
-    _Py_RefTotal--;
-    assert(v->ob_type);
-#else
-    Py_INCREF(v);
-#endif
-    assert(v->ob_refcnt == 1);
+    Py_INCREF(dead_pers_obj);
+    assert(dead_pers_obj->ob_refcnt == 1);
     /* Incremement the refcount again, because delitem is going to
-        DECREF it.  If it's refcount reached zero again, we'd call back to
+        DECREF it.  If its refcount reached zero again, we'd call back to
         the dealloc function that called us.
     */
-    Py_INCREF(v);
+    Py_INCREF(dead_pers_obj);
 
-    /* TODO:  Should we call _Py_ForgetReference() on error exit? */
     if (PyDict_DelItem(self->data, oid) < 0)
-        return;
-    Py_DECREF((ccobject *)((cPersistentObject *)v)->cache);
-    ((cPersistentObject *)v)->cache = NULL;
+    {
+        /* Almost ignore errors if it wasn't already present (somehow;
+           that shouldn't be possible since we literally just got it out
+           of this dict and we're holding the GIL and not making any
+           calls that could cause a greenlet switch so the state of the
+           dictionary should not change). We still need to finish the cleanup.
 
-    assert(v->ob_refcnt == 1);
+           Just write an unraisable error (like an exception from __del__,
+           because that's basically what this is).
+        */
+        PyErr_WriteUnraisable((PyObject*)dead_pers_obj);
+        PyErr_Clear();
+        /* Have the same side effect of clearing a ref count as the dict would have.*/
+        Py_DECREF(dead_pers_obj);
+    }
+    /* Now remove the dead object's reference to self. Note that this could
+       cause self to be dealloced.
+    */
+    Py_DECREF(dead_pers_obj_ref_to_self);
+    dead_pers_obj->cache = NULL;
 
-    /* Undo the temporary resurrection.
-        Don't DECREF the object, because this function is called from
-        the object's dealloc function. If the refcnt reaches zero, it
+    assert(dead_pers_obj->ob_refcnt == 1);
+
+    /* Don't DECREF the object, because this function is called from
+        the object's dealloc function. If the refcnt reaches zero (again), it
         will all be invoked recursively.
     */
-    _Py_ForgetReference(v);
 }
 
 static PyObject *
@@ -711,13 +726,12 @@ cc_new_ghost(ccobject *self, PyObject *args)
     {
         /* Its a persistent class, such as a ZClass. Thats ok. */
     }
-    else if (v->ob_type->tp_basicsize < sizeof(cPersistentObject))
+    else if (! PER_TypeCheck(v))
     {
         /* If it's not an instance of a persistent class, (ie Python
             classes that derive from persistent.Persistent, BTrees,
             etc), report an error.
 
-            TODO:  checking sizeof() seems a poor test.
         */
         PyErr_SetString(PyExc_TypeError,
                         "Cache values must be persistent objects.");
@@ -733,7 +747,7 @@ cc_new_ghost(ccobject *self, PyObject *args)
     Py_DECREF(tmp);
     if (tmp != Py_None)
     {
-        PyErr_SetString(PyExc_AssertionError,
+        PyErr_SetString(PyExc_ValueError,
                         "New ghost object must not have an oid");
         return NULL;
     }
@@ -745,7 +759,7 @@ cc_new_ghost(ccobject *self, PyObject *args)
     Py_DECREF(tmp);
     if (tmp != Py_None)
     {
-        PyErr_SetString(PyExc_AssertionError,
+        PyErr_SetString(PyExc_ValueError,
                         "New ghost object must not have a jar");
         return NULL;
     }
@@ -754,7 +768,7 @@ cc_new_ghost(ccobject *self, PyObject *args)
     if (tmp)
     {
         Py_DECREF(tmp);
-        PyErr_SetString(PyExc_AssertionError,
+        PyErr_SetString(PyExc_ValueError,
                         "The given oid is already in the cache");
         return NULL;
     }
@@ -891,6 +905,7 @@ cc_init(ccobject *self, PyObject *args, PyObject *kwds)
 static void
 cc_dealloc(ccobject *self)
 {
+    PyObject_GC_UnTrack((PyObject *)self);
     Py_XDECREF(self->data);
     Py_XDECREF(self->jar);
     PyObject_GC_Del(self);
@@ -1033,13 +1048,11 @@ cc_add_item(ccobject *self, PyObject *key, PyObject *v)
     {
         /* Its a persistent class, such as a ZClass. Thats ok. */
     }
-    else if (v->ob_type->tp_basicsize < sizeof(cPersistentObject))
+    else if (! PER_TypeCheck(v))
     {
         /* If it's not an instance of a persistent class, (ie Python
             classes that derive from persistent.Persistent, BTrees,
             etc), report an error.
-
-            TODO:  checking sizeof() seems a poor test.
         */
         PyErr_SetString(PyExc_TypeError,
                         "Cache values must be persistent objects.");
@@ -1058,7 +1071,7 @@ cc_add_item(ccobject *self, PyObject *key, PyObject *v)
         PyErr_Format(PyExc_TypeError,
                     "Cached object oid must be bytes, not a %s",
                     oid->ob_type->tp_name);
-        
+
         return -1;
     }
 
@@ -1307,7 +1320,6 @@ static PyTypeObject Cctype =
     (initproc)cc_init,                  /* tp_init */
 };
 
-#ifdef PY3K
 static struct PyModuleDef moduledef =
 {
     PyModuleDef_HEAD_INIT,
@@ -1321,39 +1333,24 @@ static struct PyModuleDef moduledef =
     NULL,                               /* m_free */
 };
 
-#endif
-
 static PyObject*
 module_init(void)
 {
   PyObject *module;
 
-#ifdef PY3K
     ((PyObject*)&Cctype)->ob_type = &PyType_Type;
-#else
-    Cctype.ob_type = &PyType_Type;
-#endif
     Cctype.tp_new = &PyType_GenericNew;
     if (PyType_Ready(&Cctype) < 0)
     {
         return NULL;
     }
 
-#ifdef PY3K
     module = PyModule_Create(&moduledef);
-#else
-    module = Py_InitModule3("cPickleCache", NULL, cPickleCache_doc_string);
-#endif
 
-#ifdef PY3K
-    capi = (cPersistenceCAPIstruct *)PyCapsule_Import(CAPI_CAPSULE_NAME, 0);
-#else
-    capi = (cPersistenceCAPIstruct *)PyCObject_Import(
-                    "persistent.cPersistence", "CAPI");
-#endif
-    if (!capi)
+    cPersistenceCAPI = (cPersistenceCAPIstruct *)PyCapsule_Import(CAPI_CAPSULE_NAME, 0);
+    if (!cPersistenceCAPI)
         return NULL;
-    capi->percachedel = (percachedelfunc)cc_oid_unreferenced;
+    cPersistenceCAPI->percachedel = (percachedelfunc)cc_oid_unreferenced;
 
     py__p_changed = INTERN("_p_changed");
     if (!py__p_changed)
@@ -1378,14 +1375,7 @@ module_init(void)
     return module;
 }
 
-#ifdef PY3K
 PyMODINIT_FUNC PyInit_cPickleCache(void)
 {
     return module_init();
 }
-#else
-PyMODINIT_FUNC initcPickleCache(void)
-{
-    module_init();
-}
-#endif
